@@ -1,13 +1,16 @@
 #define PY_SSIZE_T_CLEAN
+// https://stackoverflow.com/questions/39760887/warning-message-when-including-numpy-arryobject-h
+//#define NPY_NO_DEPRECATED_API 
 #include <Python.h>                // Must be first
 #include <numpy/arrayobject.h>
 #include <stdio.h>
 #include <math.h>
 #include <stdbool.h>
+#include <stdint.h>
 
 
 static char module_docstring[] = "\
-Module computes the Canada Fire Index on a gridded dataset.";
+Canadafire 1.1.0 module computes the Canada Fire Index on a gridded dataset.";
 
 // For help with Python extension:
 // https://stackoverflow.com/questions/56182259/how-does-one-acces-numpy-multidimensionnal-array-in-c-extensions
@@ -28,6 +31,7 @@ static inline double ipow8(double base)
     return base8;
 }
 
+#if 0
 /** Computes exponentiation using repeated squares: base^exp */
 static inline double ipow(double base, int exp)
 {
@@ -41,12 +45,20 @@ static inline double ipow(double base, int exp)
 
     return result;
 }
+#endif
 
 /** Checks that arr has same rank and dimensions as arr0 */
 bool check_rank_dims(char const *name0, PyArrayObject *arr0, char const *name, PyArrayObject *arr)
 {
     char msg[256];
 
+    // Check type
+    if (PyArray_DESCR(arr)->type_num != NPY_DOUBLE) {
+        PyErr_SetString(PyExc_TypeError, "Input parameters for canadafire must have type double");
+        return NULL;
+    }
+
+
     // Check rank
     if (PyArray_NDIM(arr) != PyArray_NDIM(arr0)) {
         sprintf(msg, "Parameter %s must have same rank as %s", name, name0);
@@ -140,7 +152,7 @@ double *_bui, double *_ffm, double *_isi, double *_fwi, double *_dsr, double *_d
         if(mo < ew) {
             // EQ 7a
             double const kl =
-                0.424 * (1.- pow((100.-h)/100, 1.7))
+                0.424 * (1.- pow((100.-h)/100., 1.7))
                 + (0.0694 * sqrt(w)) * (1. - ipow8((100.-h)/100.));
 
             // EQ 7b
@@ -318,6 +330,9 @@ double *_bui, double *_ffm, double *_isi, double *_fwi, double *_dsr, double *_d
     *_dmc = dmc;
     *_dc = dc;
 
+    // DEBUGGING:
+    // printf("%f %f %f %f | %f %f %f %f %f %f %f %f\n", t,h,w,r,  u,m,ffm,isi,fwi,dsr,dmc,dc);
+
 }
 
 // ===================================================================
@@ -428,30 +443,35 @@ static PyObject* canadafire_canadafire(PyObject *module, PyObject *args, PyObjec
     double ffmc00 = 85.5;    // Initial fine fuel moisture code (FMC)
     double dmc00 = 6.0;      // Initial duff moister code (DMC)
     double dc00 = 15.0;      // Initial drought code (DC)
+    PyArrayObject *mask_out = NULL;
+    double fill_value = -1.e10;    // Value for masked-out gridcells
 
     // List must include ALL arg names, including positional args
     static char *kwlist[] = {
         "tin", "hin", "win", "rin",    // *args
         "imonth", "ffmc0", "dmc0", "dc0", "debug",         // **kwargs
+        "mask_out", "fill_value",
         NULL};
     // -------------------------------------------------------------------------    
     /* Parse the Python arguments into Numpy arrays */
     // p = "predicate" for bool: https://stackoverflow.com/questions/9316179/what-is-the-correct-way-to-pass-a-boolean-to-a-python-c-extension
-    if(!PyArg_ParseTupleAndKeywords(args, kwargs, "O!O!O!O!|O!dddp",
+    if(!PyArg_ParseTupleAndKeywords(args, kwargs, "O!O!O!O!|O!dddpO!d",
         kwlist,
         &PyArray_Type, &tin,    // tin(yx,t)
         &PyArray_Type, &hin,    // hin(yx,t)
         &PyArray_Type, &win,    // win(yx,t)
         &PyArray_Type, &rin,    // rin(yx,t)
 //        &PyArray_Type, &imonth,    // imonth(t)
-        &PyArray_Type, &imonth, &ffmc00, &dmc00, &dc00, &debug
+        &PyArray_Type, &imonth, &ffmc00, &dmc00, &dc00, &debug,
+        &PyArray_Type, &mask_out, &fill_value
         )) return NULL;
 
     // Construct standard 183-day imonth array if none given.
+    bool imonth_alloc = false;
     if (imonth == NULL) {
-        static const int imonth_dims[] = {183};
-        static const int imonth_strides[] = {sizeof(int)};
-        static const int imonth_data[] = {
+        static npy_intp imonth_dims[] = {183};
+        static npy_intp imonth_strides[] = {sizeof(int)};
+        static int imonth_data[] = {
             4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,      // April
             5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,    // May
             6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,      // June
@@ -461,6 +481,7 @@ static PyObject* canadafire_canadafire(PyObject *module, PyObject *args, PyObjec
         imonth = (PyArrayObject*) PyArray_NewFromDescr(&PyArray_Type, 
             PyArray_DescrFromType(NPY_INT), 1, imonth_dims, imonth_strides, imonth_data,
             NPY_ARRAY_C_CONTIGUOUS | NPY_ARRAY_ALIGNED, NULL);
+        imonth_alloc = true;
     }
 
 
@@ -484,7 +505,7 @@ static PyObject* canadafire_canadafire(PyObject *module, PyObject *args, PyObjec
     // -----------------------------------------
     // Make sure main arrays all have same rank and dimensions
     for (int i=0; i<ninputs; ++i) {
-        if (!check_rank_dims(input_names[i], inputs0[i], input_names[0], inputs0[0]))
+        if (!check_rank_dims(input_names[0], inputs0[0], input_names[i], inputs0[i]))
             return NULL;
     }
 
@@ -504,6 +525,7 @@ static PyObject* canadafire_canadafire(PyObject *module, PyObject *args, PyObjec
         for (int j=0; j<ndim0-1; ++j) nxy *= PyArray_DIM(inputs0[0],j);
         npy_intp _dims1[] = {nxy, ntime};
         PyArray_Dims shape1 = {_dims1, 2};
+        PyArray_Dims shape1b = {_dims1, 1};
 
         // This will copy arrays if not already in C order.
         // That would be a good thing, since it would make time the lowest-stride dimension.
@@ -515,6 +537,11 @@ static PyObject* canadafire_canadafire(PyObject *module, PyObject *args, PyObjec
         if (!win) return NULL;
         rin = (PyArrayObject *)PyArray_Newshape(rin, &shape1, NPY_CORDER);
         if (!rin) return NULL;
+
+        if (mask_out != NULL) {
+            mask_out = (PyArrayObject *)PyArray_Newshape(mask_out, &shape1b, NPY_CORDER);
+            if (!mask_out) return NULL;
+        }
     }
 
     // -------------------------------------------------
@@ -539,9 +566,25 @@ static PyObject* canadafire_canadafire(PyObject *module, PyObject *args, PyObjec
         return NULL;
     }
 
+    if (mask_out != NULL) {
+        if (PyArray_DESCR(mask_out)->type_num != NPY_BOOL) {
+            PyErr_SetString(PyExc_TypeError, "Parameter mask_out must have type bool");
+            return NULL;
+        }
+        if (PyArray_NDIM(mask_out) != 1) {
+            PyErr_SetString(PyExc_TypeError,
+                "Parameter mask_out must have just spatial dimension(s), no time");
+            return NULL;
+        }
+        if (PyArray_DIM(mask_out,0) != nxy) {
+            sprintf(msg, "Parameter mask_out must have total size %d equal to other variables; but it has %d instead.", (int)nxy, (int)PyArray_DIM(mask_out,0));
+        }
+    }
+
+
     // -------------------------------------------------------------------------    
 
-    // Output arrays
+    // Output arrays (1 spatial dimension)
     PyArrayObject *ffmcout = (PyArrayObject *)PyArray_NewLikeArray(tin, NPY_ANYORDER, NULL, 0);
     PyArrayObject *isiout = (PyArrayObject *) PyArray_NewLikeArray(tin, NPY_ANYORDER, NULL, 0);
     PyArrayObject *fwiout = (PyArrayObject *)PyArray_NewLikeArray(tin, NPY_ANYORDER, NULL, 0);
@@ -573,13 +616,28 @@ static PyObject* canadafire_canadafire(PyObject *module, PyObject *args, PyObjec
             const double w = *((double *)PyArray_GETPTR2(win,ii,ti));
             const double r = *((double *)PyArray_GETPTR2(rin,ii,ti));
             const int im =  *((int *)PyArray_GETPTR1(imonth,ti));
+            // If no mask_out, compute for all pixels
+            const char msk_out = (mask_out == NULL ? 0 :
+                *((char *)PyArray_GETPTR1(mask_out,ii)));
 
             // Run the core computation on a single gridpoint.
             double bui, ffm, isi, fwi, dsr, dmc, dc;   // Output vars
-            canadafire(
-                t, h, w, r, im,
-                ffmc0, dmc0, dc0,    // Running values from one day to the next
-                &bui, &ffm, &isi, &fwi, &dsr, &dmc, &dc);
+            if (msk_out) {
+                bui = fill_value;
+                ffm = fill_value;
+                isi = fill_value;
+                fwi = fill_value;
+                dmc = fill_value;
+                dc = fill_value;
+            } else {
+                // Keep humidity in range [0,100]
+            	double _h = (h>100. ? 100. : h);
+            	_h = (_h < 0. ? 0. : _h);
+                canadafire(
+                    t, _h, w, r, im,
+                    ffmc0, dmc0, dc0,    // Running values from one day to the next
+                    &bui, &ffm, &isi, &fwi, &dsr, &dmc, &dc);
+            }
 
             // housekeeping items
             // set todays values to the yesterdays values before going on
@@ -599,30 +657,234 @@ static PyObject* canadafire_canadafire(PyObject *module, PyObject *args, PyObjec
     }
 
     // ---------------------------------------------------------
-    // Reshape to original rank
-    ffmcout = (PyArrayObject *)PyArray_Newshape(ffmcout, &shape0, NPY_ANYORDER);
+    // Free input arrays we allocated (the version w/ 1 spatial dimension)
+    Py_DECREF(tin);
+    Py_DECREF(hin);
+    Py_DECREF(win);
+    Py_DECREF(rin);
+    if (imonth_alloc) Py_DECREF(imonth);
+
+    // ---------------------------------------------------------
+    // Reshape to original rank; and free the 1-spatial-dim version
+    PyArrayObject *_ffmcout = (PyArrayObject *)PyArray_Newshape(ffmcout, &shape0, NPY_ANYORDER);
     if (!ffmcout) return NULL;
-    isiout = (PyArrayObject *)PyArray_Newshape(isiout, &shape0, NPY_ANYORDER);
+    Py_DECREF(ffmcout);
+    PyArrayObject *_isiout = (PyArrayObject *)PyArray_Newshape(isiout, &shape0, NPY_ANYORDER);
     if (!isiout) return NULL;
-    fwiout = (PyArrayObject *)PyArray_Newshape(fwiout, &shape0, NPY_ANYORDER);
+    Py_DECREF(isiout);
+    PyArrayObject *_fwiout = (PyArrayObject *)PyArray_Newshape(fwiout, &shape0, NPY_ANYORDER);
     if (!fwiout) return NULL;
-    dmcout = (PyArrayObject *)PyArray_Newshape(dmcout, &shape0, NPY_ANYORDER);
+    Py_DECREF(fwiout);
+    PyArrayObject *_dmcout = (PyArrayObject *)PyArray_Newshape(dmcout, &shape0, NPY_ANYORDER);
     if (!dmcout) return NULL;
-    buiout = (PyArrayObject *)PyArray_Newshape(buiout, &shape0, NPY_ANYORDER);
+    Py_DECREF(dmcout);
+    PyArrayObject *_buiout = (PyArrayObject *)PyArray_Newshape(buiout, &shape0, NPY_ANYORDER);
     if (!buiout) return NULL;
-    dcout = (PyArrayObject *)PyArray_Newshape(dcout, &shape0, NPY_ANYORDER);
+    Py_DECREF(buiout);
+    PyArrayObject *_dcout = (PyArrayObject *)PyArray_Newshape(dcout, &shape0, NPY_ANYORDER);
     if (!dcout) return NULL;
+    Py_DECREF(dcout);
 
     // Return a tuple of the output arrays we created.
     // https://stackoverflow.com/questions/3498210/returning-a-tuple-of-multipe-objects-in-python-c-api
-    return PyTuple_Pack(6, buiout,ffmcout,isiout,fwiout,dmcout,dcout);
+    PyObject *ret = PyTuple_Pack(6, _buiout,_ffmcout,_isiout,_fwiout,_dmcout,_dcout);
+    return ret;
 }
+// ============================================================
+// https://stackoverflow.com/questions/11498169/dealing-with-angle-wrap-in-c-code
+static inline double constrain_range(double const x, double const range)
+{
+    double y = fmod(x,range);
+    if (y < 0) return y + range;
+    return y;
+}
+
+inline bool is_leap_year(int year)
+{
+    return (year % 4) == 0 && ((year % 100) != 0 || (year % 400) == 0);
+}
+
+/* Approximate equation of time.
+This is accurate to about 1-2 minutes of time.  One can compare its results against:
+     https://gml.noaa.gov/grad/solcalc/
+
+https://susdesign.com/popups/sunangle/eot.php
+
+For example, the EOT adjustment in mid-February is about -14
+minutes. So when converting clock time to local solar time, you'd
+subtract 14 minutes. When converting from local solar time to clock
+time, you'd add 14 minutes.
+
+The EOT can be approximated by the following formula:
+   E = 9.87 * sin (2B) - 7.53 * cos (B) - 1.5 * sin (B)
+
+Where:
+   B = 360 * (N - 81) / 365
+
+Where:
+   N = day number, January 1 = day 1
+
+Note: The SunAngle program currently uses a more sophisticated
+algorithm for EOT calculations, but the above formula is a decent
+approximation and much simpler. Note also that the EOT output is in
+hours, so please multiply by 60 if you'd like to obtain results in
+minutes.
+
+*/
+inline double equation_of_time(int const Y, int const doy, double const UT_h)
+{
+
+    // Adjust doy for time of day
+    double const N = (double)doy + (double)UT_h / 24.;
+
+    // Simple / Uncorrected Equation of Time
+    // https://susdesign.com/popups/sunangle/eot.ph
+    double const B = 2.*M_PI * (N - 81.) / (is_leap_year(Y) ? 366. : 365.);    // [Radians]
+    double const E = (9.87 * sin(2.*B) - 7.53*cos(B) - 1.5*sin(B)) / 60.;    // [hours]
+
+    return E;
+}
+
+
+
+static inline void solar_noon(int const Y, int const doy, double longitude_deg,
+    double *mean_solar_noon,        // OUT (hours)
+    double *apparent_solar_noon)    // OUT (hours)
+// longitude: [-180,180]
+{
+    // Obtain UTC for mean solar noon at this longitude
+    // We will compute the equation of time at this time.
+    double const UT = (720. - 4 * longitude_deg) / 1440.; // [0.0, 1.0]
+    double const UT_norm = (UT >=0 ? UT : UT + 1);
+    double const UT_h = UT_norm * 24.;    // Time of day (hours)
+//double const UT_h = 12.;
+
+    // The precise definition of the Equation of Time is
+    // E = GHA (apparent Sun) - GHA (mean Sun)
+    double const E_h = equation_of_time(Y,doy, UT_h);
+//    printf("E_h = %f\n", E_h);
+
+
+    *mean_solar_noon = UT_h;
+    *apparent_solar_noon = constrain_range(UT_h - E_h, 24.0);
+    
+
+}
+    
+
+static char canadafire_solar_noon_docstring[] = "\
+Usage:\n\
+    msn,asn = solar_noon(Y,M,D, lons)\n\
+\n\
+    Y,M,D: int\n\
+        Date to compute solar noon\n\
+    lons: array(...)\n\
+        Longitudes at which to compute solar noon on the given date.\n\
+    msn: array(...)\n\
+        For each longitude, time (UTC) of mean solar noon\n\
+        (i.e. based only on longitude)\n\
+    asn: array(...)\n\
+        For eacn longitude, time (in UTC) sun is highest in the sky";
+
+// https://numpy.org/devdocs/user/c-info.ufunc-tutorial.html
+static PyObject *canadafire_solar_noon(PyObject *module, PyObject *args, PyObject *kwargs)
+{
+    int Y,M,D;
+    int doy=-1;
+    PyArrayObject *lons;    // Longitudes over which to operate
+
+
+    // List must include ALL arg names, including positional args
+    static char *kwlist[] = {
+        "Y", "M", "D", "lons",    // *args
+        "doy",        // **kwargs
+        NULL};
+    // -------------------------------------------------------------------------    
+    /* Parse the Python arguments into Numpy arrays */
+    // p = "predicate" for bool: https://stackoverflow.com/questions/9316179/what-is-the-correct-way-to-pass-a-boolean-to-a-python-c-extension
+    if(!PyArg_ParseTupleAndKeywords(args, kwargs, "iiiO!|i",
+        kwlist,
+        &Y, &M, &D,
+        &PyArray_Type, &lons,    // lons(...)
+        &doy
+        )) return NULL;
+
+    // Check type
+    if (PyArray_DESCR(lons)->type_num != NPY_DOUBLE) {
+        PyErr_SetString(PyExc_TypeError, "Parameter lons must have type double");
+        return NULL;
+    }
+
+    // Use doy if it is supplied; otherwise get doy from Y/M/D
+    // https://stackoverflow.com/questions/19110675/calculating-day-of-year-from-date/19110801
+    if (doy < 0) {
+        struct tm date = {};
+        date.tm_year = Y - 1900;
+        date.tm_mon = M - 1;
+        date.tm_mday = D;
+        mktime( &date );
+        doy = date.tm_yday + 1;
+    }
+
+
+    // ------- Allocate output arrays
+    // Mean solar noon
+    PyArrayObject *msn = (PyArrayObject *)PyArray_NewLikeArray(lons, NPY_ANYORDER, NULL, 0);
+    if (msn == NULL) return NULL;
+    // Apparent solar noon
+    PyArrayObject *asn = (PyArrayObject *)PyArray_NewLikeArray(lons, NPY_ANYORDER, NULL, 0);
+    if (asn == NULL) return NULL;
+
+
+    // Run the calculation
+    // Array Iterators:
+    //    https://numpy.org/devdocs/user/c-info.beyond-basics.html#iterating-over-elements-in-the-array
+    // Reference counting:
+    //    https://pythonextensionpatterns.readthedocs.io/en/latest/refcount.html
+    // Numpy Array API:
+    //    http://pageperso.lif.univ-mrs.fr/~francois.denis/IAAM1/numpy-html-1.14.0/reference/c-api.array.html
+
+    PyArrayIterObject *loni, *msni, *asni;
+    loni = (PyArrayIterObject *)PyArray_IterNew((PyObject *)lons);
+    if (loni == NULL) return NULL;
+    msni = (PyArrayIterObject *)PyArray_IterNew((PyObject *)msn);
+    if (msni == NULL) return NULL;
+    asni = (PyArrayIterObject *)PyArray_IterNew((PyObject *)asn);
+    if (asni == NULL) return NULL;
+    while (loni->index < loni->size) {
+        const double lon = *((double *) PyArray_ITER_DATA(loni));
+
+        // Stores result in msn (mean solar noon) and asn (apparent solar noon).
+        solar_noon(Y,doy,lon,
+            (double *) PyArray_ITER_DATA(msni),
+            (double *) PyArray_ITER_DATA(asni));
+
+        PyArray_ITER_NEXT(loni);
+        PyArray_ITER_NEXT(msni);
+        PyArray_ITER_NEXT(asni);
+    }
+
+    Py_DECREF(asni);
+    Py_DECREF(msni);
+    Py_DECREF(loni);
+
+    return PyTuple_Pack(2, msn, asn);
+}
+
+
+
 // ============================================================
 // Random other Python C Extension Stuff
 static PyMethodDef CanadafireMethods[] = {
     {"canadafire",
-        canadafire_canadafire,
+        (PyCFunction)canadafire_canadafire,
         METH_VARARGS | METH_KEYWORDS, canadafire_canadafire_docstring},
+
+    {"solar_noon",
+        (PyCFunction)canadafire_solar_noon,
+        METH_VARARGS | METH_KEYWORDS, canadafire_solar_noon_docstring},
+
+    // Sentinel
     {NULL, NULL, 0, NULL}
 };
 
